import click
from ruamel.yaml import YAML

from backups_rotate.area import Area

DEFAULT_CONFIG_FILE = "/etc/backups_rotate.yaml"

@click.command()
@click.option(
    "--config",
    "-c",
    default=DEFAULT_CONFIG_FILE,
    help="Configuration file",
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    help="Verbose mode",
)
@click.option(
    "--dry-run",
    "-n",
    is_flag=True,
    help="Dry run",
)
@click.argument('area_names', nargs=-1)
def cli(config, verbose, dry_run, area_names):
    """
    backups_rotate: a tool to manage backups
    """

    # read the configuration
    yaml = YAML(typ='safe')
    with open(config) as f:
        config = yaml.load(f)
        areas_by_name = {area["name"]: area for area in config}

    if not area_names:
        area_dicts = config
    else:
        try:
            area_dicts = [areas_by_name[name] for name in area_names]
        except KeyError as e:
            click.echo(f"Area {e} not found in configuration")
            return 1

    areas = [Area(area_dict) for area_dict in area_dicts]

    if verbose:
        for area in areas:
            area.read()
            if verbose:
                click.echo(f"area={area}")
            area.delete(dry_run=dry_run, verbose=verbose)


if __name__ == "__main__":
    cli()



"""
an area has a name, and a folder,
and it refers to a set of files defined by some patterns
together with a policy
"""

from pathlib import Path
from dataclasses import dataclass
from datetime import datetime
from collections.abc import Iterator

import pandas as pd

from .policy import Policy


def extract_date_from_filename(path: Path, dateformat) -> pd.Timestamp | None:
    """
    Extracts a date from a filename using the provided date format.

    Args:
        filename (str): The filename to search for a date.
        dateformat (str): A strptime-compliant date format to match.

    Returns:
        datetime: The extracted date if found, None otherwise.
    """
    filename = path.name
    # we need to know how many characters to try and match
    dummy = datetime(2000, 1, 1, 0, 0, 0)
    partial_length = len(dummy.strftime(dateformat))

    for i in range(len(filename) - partial_length + 1):
        substring = filename[i:i + partial_length]
        try:
            # print(substring)
            return pd.Timestamp(datetime.strptime(substring, dateformat))
        except ValueError:
            continue
    return None


@dataclass
class TimedPath:
    """
    a path with a timestamp
    knows how to get the timestamp from the path or a format
    """
    path: Path
    timestamp: pd.Timestamp = None

    def attach_timestamp(self, *, datetime_format: str = None,
                         use_modification_time: bool = False):
        """
        attach a timestamp to the path

        if datetime_format is provided, it is used to parse the timestamp from
        the filename

        otherwise, and if use_modification_time is True, the modification time
        is used - the default is to use the creation time
        """
        # no format
        if datetime_format is None:
            if use_modification_time:
                self.timestamp = pd.Timestamp(self.path.stat().st_mtime)
            else:
                self.timestamp = pd.Timestamp(self.path.stat().st_ctime)
            return
        # otherwise use the format
        date = extract_date_from_filename(self.path, datetime_format)
        if date is None:
            raise ValueError(
                f"Could not extract date from {self.path} using {datetime_format}")
        self.timestamp = date

    def __lt__(self, other):
        return self.timestamp < other.timestamp


class Area:
    """
    an area is a set of files with a policy
    """

    def __init__(self, area_dict: dict):
        try:
            self.name = area_dict['name']
            self.path = Path(area_dict['folder'])
            self.patterns = area_dict['patterns']
            self.policy = Policy(area_dict['policy'])
        except KeyError as e:
            raise ValueError(f"Missing key {e}") from e
        self.datetime_format = area_dict.get('datetime_format', None)
        self.use_modification_time = area_dict.get('use_modification_time', None)
        # for the record
        self.area_dict = area_dict
        # stats
        self.total = 0
        self.deleted = 0
        # list of TimedPath
        self._files = None
        self._kept = None
        # sanity check
        self._check()

    def __repr__(self):
        return f"Area({self.name} on {self.path} - {self.total} files, {self.deleted} deleted)"

    def _check(self):
        """
        sanity check
        """
        # no need to check for mandatory fields
        # as they are used in the constructor already
        allowed = set("name|folder|patterns|policy|datetime_format|use_modification_time".split("|"))
        # any non-supported key ? this can be dangerous
        unsupported = set(self.area_dict.keys()) - allowed
        if unsupported:
            raise ValueError(f"Invalid keys in {self.name} : {unsupported}")

        if not isinstance(self.patterns, list):
            raise ValueError(f"Invalid patterns in {self.name} - should be a list")
        if not isinstance(self.use_modification_time, (bool, type(None))):
            raise ValueError(f"Invalid use_modification_time in {self.name} - should be a boolean")
        if self.datetime_format is not None and not isinstance(self.datetime_format, str):
            raise ValueError(f"Invalid datetime_format in {self.name} - should be a string")
        if self.use_modification_time is not None and self.datetime_format is not None:
            raise ValueError(f"Cannot have both use_modification_time and datetime_format {self.name}")

        # check the folder
        if not self.path.exists():
            raise ValueError(f"Folder {self.path} does not exist")
        if not self.path.is_dir():
            raise ValueError(f"{self.path} is not a folder")
        if not self.path.is_absolute():
            raise ValueError(f"{self.path} is not an absolute path")


    def _populate(self):
        """
        read the disk to populate the list of files
        """
        self._files = []
        for pattern in self.patterns:
            self._files.extend(self.path.glob(pattern))
        self._files = [TimedPath(path) for path in self._files]
        self.total = len(self._files)

    def _time_files(self):
        """
        attach a timestamp to each file
        """
        for file in self._files:
            file.attach_timestamp(
                datetime_format=self.datetime_format,
                use_modification_time=self.use_modification_time)
        # use TimedPath __lt__ to sort by timestamp
        self._files.sort()

    def read(self):
        """
        read the disk and computes which files to keep
        """
        self._populate()
        self._time_files()
        self._kept = self.policy.keep_timestamps([
            file.timestamp for file in self._files])
        self.deleted = 0
        for i, file in enumerate(self._files):
            if not self._kept[i]:
                self.deleted += 1


    def _iterate(self) -> Iterator[tuple[bool, Path, pd.Timestamp]]:
        """
        iterate on tuples (path: Path, kept: bool)
        """
        if self._kept is None:
            print("warning: area not read yet - bailing out")
            return
        for i, file in enumerate(self._files):
            yield (self._kept[i], file.path, file.timestamp)


    def list(self, *, deleted=True, kept=True):
        """
        shows the files in the area, with an indication of kept/deleted
        """
        if self._files is None:
            print("warning: area not populated yet")
            return
        if self._kept is None:
            print("warning: area not read yet - showing all files")
            for file in self._files:
                print(file)
            return
        for kept_file, file, ts in self._iterate():
            # show only selected entries
            if (deleted and not kept_file) or (kept and kept_file):
                print("KEEP" if kept_file else "DELE", end=" ")
                print(file, ts)

    def delete(self, *, dry_run=True, verbose=False):
        """
        delete the files in the area
        """
        for kept_file, file, _ts in self._iterate():
            if kept_file:
                if dry_run:
                    print("dry-run: would KEEP  ", file)
                continue
            if dry_run:
                print("dry-run: would DELETE", file)
            elif not file.exists():
                print("MISSING", file)
            else:
                if verbose:
                    print("DELETING", file)
                file.unlink()



"""
the Policy class is in charge of spotting the timestamps that are to be kept
"""


import numpy as np
import pandas as pd


# the periods are hard-coded for now
PERIODS = {
    'year': '1YE',
    'quarter': '1QE',
    'month': '1ME',
    'week': '1W',
    'day': '1D',
    'hour': '1h',
}


class Policy:

    """
    A policy is a set of rules that determine which timestamps are kept
    and which are discarded.
    It is definded by a dictionary that may have the following keys:
    'year', 'quarter', 'month', 'week', 'day', 'hour'
    with values that are either integers or the string 'infinite'.
    """

    def __init__(self, period_dict: dict):
        self.period_dict = period_dict
        for k in period_dict.keys():
            if k not in PERIODS:
                raise ValueError(f"Invalid period: {k}")

    def __repr__(self):
        return f"Policy({self.period_dict})"


    def count_validating_periods_for_timestamps(
            self, timestamps: list[pd.Timestamp]) -> np.ndarray:

        """
        Apply the policy to a list of timestamps
        returns a ndarray of size len(timestamps) and of type uint8
        that indicates which timestamps are to be kept

        specifically, the value indicates for how many period types
        the timestamp qualifies to be kept
        """

        total = len(timestamps)
        kept = np.zeros(total, dtype=np.uint8)

        index = pd.DatetimeIndex(timestamps)
        rank = range(total)
        df = pd.DataFrame(rank, index=index, columns=['rank'])

        df.sort_index(inplace=True)

        for period, value in self.period_dict.items():

            # isolate the last item per period
            last_per_period = df.resample(PERIODS[period]).last()
            # skip periods without any item
            last_per_period.dropna(inplace=True)
            # keep only the last <value> items if not infinite
            if value != 'infinite':
                if value < 1:
                    raise ValueError(
                        f"Invalid value for {period}: {value} - should be >= 1")
                last_per_period = last_per_period.iloc[-value:]
            last_per_period = last_per_period.astype(int)
            # print("===", period)
            # for rank in last_per_period['rank']:
            #     print(rank, timestamps[rank])
            # mark the items to be kept
            kept[last_per_period['rank']] += 1

        return kept

    def keep_timestamps(self, timestamps: list[pd.Timestamp]) -> np.ndarray:
        """
        a wrapper around count_validating_periods_for_timestamps
        that returns a boolean array instead
        """
        return self.count_validating_periods_for_timestamps(timestamps) > 0

    def indices_of_timestamps_to_keep(self, timestamps: list[pd.Timestamp]) -> list[int]:
        """
        return the indices of the timestamps to keep
        """
        return np.nonzero(self.keep_timestamps(timestamps))[0].tolist()



# pylint: disable=missing-docstring, wildcard-import, multiple-statements

# from unittest import TestCase

from backups_rotate import Policy

from tests.samples import *
from tests.policies import *


# the generic method is to compare the output of the policy
# with the expected output
def _test_policy_from_constants(sample_index, policy_index):
    policy_dict = globals()[f'POLICY{policy_index}']
    sample = globals()[f'sample{sample_index}']()
    kept = Policy(policy_dict).keep_timestamps(sample)
    expected = globals()[f'POLICY{policy_index}_SAMPLE{sample_index}']
    assert kept.sum() == expected

def test_policy_1_1(): _test_policy_from_constants(1, 1)
def test_policy_2_1(): _test_policy_from_constants(2, 1)
def test_policy_1_2(): _test_policy_from_constants(1, 2)
def test_policy_2_2(): _test_policy_from_constants(2, 2)
def test_policy_1_3(): _test_policy_from_constants(1, 3)
def test_policy_2_3(): _test_policy_from_constants(2, 3)
def test_policy_1_4(): _test_policy_from_constants(1, 4)
def test_policy_2_4(): _test_policy_from_constants(2, 4)
def test_policy_1_10(): _test_policy_from_constants(1, 10)
def test_policy_2_10(): _test_policy_from_constants(2, 10)

def test_policy_2_11(): _test_policy_from_constants(2, 11)



# pylint: disable=missing-docstring, wildcard-import

# from unittest import TestCase
from pathlib import Path

import pandas as pd
import pytest

from backups_rotate.area import extract_date_from_filename, TimedPath, Area

from tests.areas import *
from tests.policies import *


def test_broken_areas():
    for area_dict in BROKEN_AREAS:
        with pytest.raises(ValueError):
            Area(area_dict)


def test_datetime_format():
    path = Path("database-2024-01-01-something.sql")
    datetime_format = "%Y-%m-%d"
    ts = extract_date_from_filename(path, datetime_format)
    assert ts == pd.Timestamp("2024-01-01")


def test_attach_timestamp_by_format():
    path = Path("database-2024-01-01-something.sql")
    tp = TimedPath(path)
    tp.attach_timestamp(datetime_format="%Y-%m-%d")
    assert tp.timestamp == pd.Timestamp("2024-01-01")


def test_attach_timestamp_by_ctime():
    path = TEST_FOLDER / "--some-test-file--"
    if path.exists():
        path.unlink()
    now = pd.Timestamp.now()
    path.touch()
    tp = TimedPath(path)
    tp.attach_timestamp()
    assert (tp.timestamp - now) < pd.Timedelta(milliseconds=50)


def test_attach_timestamp_mydb_area():
    files = ALL_FILES
    area_dict = MYDB_AREA
    folder = Path(area_dict['folder'])
    if not folder.exists():
        folder.mkdir()
    for file in folder.glob("*"):
        file.unlink()
    for file in files:
        path = folder / file
        path.touch()

    area = Area(area_dict)
    area.read()
    assert area.deleted == 53


def test_attach_timestamp_another_area():
    files = ALL_FILES
    area_dict = ANOTHER_AREA
    folder = Path(area_dict['folder'])
    if not folder.exists():
        folder.mkdir()
    for file in folder.glob("*"):
        file.unlink()
    for file in files:
        path = folder / file
        path.touch()

    area = Area(area_dict)
    area.read()
    # because this area uses modification time, only one file
    # survives in this test
    assert area.deleted == 61
